/*****************************************************************************/
/* Kernel SOCKS Bouncer (Kernel Patch) for 2.6.x kernels [KSB26]             */
/* (c) 2004 Paolo Ardoino <ardoino.gnu@disi.unige.it>                        */
/*****************************************************************************/
/*									     */
/* This program is free software; you can redistribute it and/or modify	     */
/* it under the terms of the GNU General Public License as published by	     */
/* the Free Software Foundation; either version 2 of the License, or	     */
/* (at your option) any later version.					     */
/* This program is distributed in the hope that it will be useful,	     */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	     */
/* GNU General Public License for more details.				     */
/*									     */
/* You should have received a copy of the GNU General Public License	     */
/* along with this program; if not, write to the Free Software		     */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*****************************************************************************/

#include <linux/in.h>
#include <linux/inet.h>
#include <linux/byteorder/generic.h>
#include <linux/slab.h>
#include <linux/preempt.h>
#include <asm/semaphore.h>

#include "ksb26.h"

#define DEBUG 1 

#define SSEND { oldfs = get_fs(); set_fs(get_ds()); scall = sys_send(fd, sbuf, sblen, 0); set_fs(oldfs); }
#define SRECV { scall = 0; oldfs = get_fs(); set_fs(get_ds()); scall = sys_recv(fd, sbuf, PAGE_SIZE, 0);set_fs(oldfs); }
#define SHSOK { if(sbuf[0] != 0x05 || sbuf[1] != 0x00) return -1; }
#define SHEADERA "\x05\x01%c"
#define SHEADERR "\x05\x01%c\x03%c%s%c%c"

asmlinkage long sys_connect_orig(int fd, struct sockaddr __user *uservaddr, int addrlen);

char ksb26 = 0;
unsigned int ksb26_nsocks = 0;
unsigned int ksb26_naddrs = 0;
struct ksb26_sockscell *ksb26_sl;
struct ksb26_addrs *ksb26_al;


static char *ksb26_ntoa(unsigned long saddr)
{
	static char buff[18];
	char *p;

	p = (char *) &saddr;
	sprintf(buff, "%d.%d.%d.%d", (*p & 255), (*(p + 1) & 255), (*(p + 2) & 255), (*(p + 3) & 255));
	return buff;
}                                                        

static int ksb26_2bounce(unsigned long addr)
{
	struct ksb26_addrs *ksb26_alaux;
	unsigned int cnt = 0;

	ksb26_alaux = (struct ksb26_addrs *)ksb26_al;
	while(cnt < ksb26_naddrs && ksb26_alaux) {
		if(ksb26_alaux->addr == addr) 
			return 1;
		ksb26_alaux = ksb26_alaux->next;
		cnt++;
	}
	return 0;
}

static int ksb26_sockshs(int fd, char *dhost, int dport)
{
	int scall;
	int sblen;
	mm_segment_t oldfs;
	char *sbuf;
  
	sbuf = (char *)kmalloc(PAGE_SIZE, GFP_KERNEL);
	sblen = sprintf(sbuf, SHEADERA, 0);
	SSEND;
	SRECV;
	SHSOK;
  	sblen = sprintf(sbuf, SHEADERR, 0, strlen(dhost), dhost, (dport >> 8) & 0xFF, dport & 0xFF);
	SSEND;
	SRECV;
	SHSOK;
	kfree(sbuf);
	return scall;
}

asmlinkage long sys_connect_ksb26(int fd, struct sockaddr __user *uservaddr, int addrlen)
{
	mm_segment_t oldfs;
	struct sockaddr_in *ksin = (struct sockaddr_in *)uservaddr;
	struct ksb26_sockscell *ksb26_slaux;
	struct ksb26_sockscell *ksb26_aux = NULL;
	unsigned long int daddr;
	unsigned short int dport;
	int err = 0, scnt = 0, icnt = 0;
	struct semaphore sem;
	
	ksb26_slaux = (struct ksb26_sockscell *)ksb26_sl;
	oldfs = get_fs();
	set_fs(get_ds());
	if (ksb26_2bounce(ksin->sin_addr.s_addr) == 1) {
#ifdef DEBUG
		printk("[%s] SYS_CONNECT_KSB26 : localhost -> %s:%d [%lu]\n", MODNAME, ksb26_ntoa(ksin->sin_addr.s_addr), ntohs(ksin->sin_port), (unsigned long)ksin->sin_addr.s_addr);
#endif
		if(ksb26_nsocks == 0) {
#ifdef DEBUG
			printk("[%s] No socks5 in list.\n", MODNAME);
#endif
			return -1;
		}
		daddr = ksin->sin_addr.s_addr;
		dport = ksin->sin_port;
		set_fs(oldfs);
		while(scnt < ksb26_nsocks) {
#ifdef DEBUG
			printk("[%s] Trying %s:%d socks5.\n", MODNAME, ksb26_slaux->host, ksb26_slaux->port);
#endif
			oldfs = get_fs();
			set_fs(get_ds());
			ksin->sin_addr.s_addr = in_aton(ksb26_slaux->host);
			ksin->sin_port = htons(ksb26_slaux->port);
			set_fs(oldfs);
			if((err = sys_connect_orig(fd, uservaddr, addrlen)) != 0) {
				if (scnt < ksb26_nsocks - 1) {
					scnt++;
					ksb26_slaux = ksb26_slaux->next;
				} else 
					return err;
			} else
				break;
		}
#ifdef DEBUG
		printk("[%s] Connected to %s:%d (first) socks5.\n", MODNAME, ksb26_slaux->host, ksb26_slaux->port);
#endif
		if((err = ksb26_sockshs(fd, ksb26_ntoa(daddr), ntohs(dport))) <= 0) {
#ifdef DEBUG
			printk("[%s] Removing %s:%d socks5 from list.\n", MODNAME, ksb26_slaux->host, ksb26_slaux->port);
#endif
			sema_init(&sem, 1);
			down_interruptible(&sem);
			if(scnt == 0) {
				ksb26_aux = ksb26_sl;
				ksb26_sl = ksb26_sl->next;
				kfree(ksb26_aux);
				ksb26_slaux = ksb26_sl;
			} else {
				for(icnt = 0; icnt < scnt - 1 && ksb26_slaux->next; icnt++)
					ksb26_slaux = ksb26_slaux->next;
				if(ksb26_slaux->next)
					ksb26_aux = ksb26_slaux->next;
				if(ksb26_slaux->next->next)
					ksb26_slaux->next = ksb26_slaux->next->next;
				if(ksb26_aux)
					kfree(ksb26_aux);
			}
			up(&sem);
			ksb26_nsocks--;
#ifdef DEBUG		
			printk("[%s] socks5 removed from list.\n", MODNAME);
			printk("[%s] Please retry to connect to host.\n", MODNAME);
#endif
			return -ECONNREFUSED;
		}
#ifdef DEBUG
		printk("[%s] Connected to destination host [%s:%d].\n", MODNAME, ksb26_ntoa(daddr), ntohs(dport));
#endif
	} else { 
		set_fs(oldfs);
		err = sys_connect_orig(fd, uservaddr, addrlen);
	}
	return err;
}

EXPORT_SYMBOL(ksb26);
EXPORT_SYMBOL(ksb26_nsocks);
EXPORT_SYMBOL(ksb26_naddrs);
EXPORT_SYMBOL(ksb26_sl);
EXPORT_SYMBOL(ksb26_al);
